# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

"""
Module contains several classes used to represent internal objects within the
interpreter.
"""
from enum import Enum, auto
from zai.env import Scope
from zai.internal_error import InternalTypeError


class ObjectType(Enum):
    """
    Enum used to represent the different types of internal objects.
    """

    NUM = auto()
    STR = auto()
    ID = auto()
    BOOL = auto()
    FUNC = auto()
    NATIVE_FUNC = auto()
    CLASS_DEF = auto()
    CLASS_INSTANCE = auto()
    CLASS_METHOD = auto()
    RETURN = auto()
    NIL = auto()
    BREAK = auto()
    CONTINUE = auto()
    ARRAY = auto()
    MODULE = auto()

    def __str__(self):
        type_to_str = {
            "NUM": "number",
            "STR": "string",
            "ID": "variable name",
            "BOOL": "boolean",
            "FUNC": "function",
            "NATIVE_FUNC": "native function",
            "CLASS_DEF": "class definition",
            "CLASS_INSTANCE": "class instance",
            "CLASS_METHOD": "class method",
            "NIL": "nil",
            "ARRAY": "array",
            "MODULE": "module namespace",
        }
        return type_to_str[self.name]


class Internal_Object:
    """
    Base class for all internal objects used in the interpreter.
    """

    def __init__(
        self,
    ):
        raise NotImplementedError()


# All atomic objects
class Nil_Object(Internal_Object):
    """
    Internal object used to represent nil/null values.
    """

    def __init__(self):
        self.obj_type = ObjectType.NIL

    def __str__(self):
        return "nil"

    def __repr__(self):
        return "NIL_OBJ"

    def __eq__(self, other):
        assert other is not None, "Other is none in __eq__ function for nil object."
        return Bool_Object(self.obj_type == other.obj_type)

    def __ne__(self, other):
        return ~(self.__eq__(other))

    def __lt__(self, other):
        raise InternalTypeError("<", self.obj_type, other.obj_type)

    def __le__(self, other):
        raise InternalTypeError("<=", self.obj_type, other.obj_type)

    def __gt__(self, other):
        raise InternalTypeError(">", self.obj_type, other.obj_type)

    def __ge__(self, other):
        raise InternalTypeError(">=", self.obj_type, other.obj_type)

    def __add__(self, other):
        raise InternalTypeError("+", self.obj_type, other.obj_type)

    def __sub__(self, other):
        raise InternalTypeError("-", self.obj_type, other.obj_type)

    def __mul__(self, other):
        raise InternalTypeError("*", self.obj_type, other.obj_type)

    def __truediv__(self, other):
        raise InternalTypeError("/", self.obj_type, other.obj_type)

    def __and__(self, other):
        return Bool_Object(bool(self) and bool(other))

    def __or__(self, other):
        return Bool_Object(bool(self) or bool(other))

    def __neg__(self):
        raise InternalTypeError("-", self.obj_type)

    def __invert__(self):
        return Bool_Object(True)

    def __bool__(self):
        return False


class Bool_Object(Internal_Object):
    """
    Internal object used to represent a return value from a function.
    """

    def __init__(self, bool_val):
        self.value = bool_val
        self.obj_type = ObjectType.BOOL

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return "BOOL_OBJ {}".format(self.value)

    def __eq__(self, other):
        assert other is not None, "Other value in bool internal object __eq__ is none"
        return Bool_Object(
            self.obj_type == other.obj_type and self.value == other.value
        )

    def __ne__(self, other):
        # Using the invert operator(~) will return a new boolean object.
        return ~(self.__eq__(other))

    def __neg__(self):
        return Bool_Object(-(self.value))

    def __invert__(self):
        return Bool_Object(not (self.value))

    def __lt__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Bool_Object(self.value < other.value)
        else:
            raise InternalTypeError("<", self.obj_type, other.obj_type)

    def __le__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Bool_Object(self.value <= other.value)
        else:
            raise InternalTypeError("<=", self.obj_type, other.obj_type)

    def __gt__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Bool_Object(self.value > other.value)
        else:
            raise InternalTypeError(">", self.obj_type, other.obj_type)

    def __ge__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Bool_Object(self.value >= other.value)
        else:
            raise InternalTypeError(">=", self.obj_type, other.obj_type)

    def __add__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Num_Object(self.value + other.value)
        else:
            raise InternalTypeError("+", self.obj_type, other.obj_type)

    def __sub__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Num_Object(self.value - other.value)
        else:
            raise InternalTypeError("-", self.obj_type, other.obj_type)

    def __mul__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Num_Object(self.value * other.value)
        else:
            raise InternalTypeError("*", self.obj_type, other.obj_type)

    def __truediv__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Num_Object(self.value / other.value)
        else:
            raise InternalTypeError("/", self.obj_type, other.obj_type)

    def __and__(self, other):
        return Bool_Object(bool(self) and bool(other))

    def __or__(self, other):
        return Bool_Object(bool(self) or bool(other))

    def __bool__(self):
        return self.value


class String_Object(Internal_Object):
    """
    Internal object used to represent strings within the interpreter.
    """

    def __init__(self, string_val):
        self.value = string_val
        self.str_len = len(string_val)
        self.obj_type = ObjectType.STR

    def __repr__(self):
        return "STR_OBJ {}".format(self.value)

    def __str__(self):
        return self.value

    def __eq__(self, other):
        assert (
            other is not None
        ), "Other variable is none in __eq__ function for string object."
        return Bool_Object(
            self.obj_type == other.obj_type and self.value == other.value
        )

    def __ne__(self, other):
        return ~self.__eq__(other)

    def __lt__(self, other):
        if other.obj_type == ObjectType.STR:
            return Bool_Object(self.value < other.value)
        else:
            raise InternalTypeError("<", self.obj_type, other.obj_type)

    def __le__(self, other):
        if other.obj_type == ObjectType.STR:
            return Bool_Object(self.value <= other.value)
        else:
            raise InternalTypeError("<=", self.obj_type, other.obj_type)

    def __gt__(self, other):
        pass
        if other.obj_type == ObjectType.STR:
            return Bool_Object(self.value > other.value)
        else:
            raise InternalTypeError(">", self.obj_type, other.obj_type)

    def __ge__(self, other):
        if other.obj_type == ObjectType.STR:
            return Bool_Object(self.value >= other.value)
        else:
            raise InternalTypeError(">=", self.obj_type, other.obj_type)

    def __add__(self, other):
        if other.obj_type == ObjectType.STR:
            return String_Object(self.value + other.value)
        else:
            raise InternalTypeError("+", self.obj_type, other.obj_type)

    def __sub__(self, other):
        raise InternalTypeError("-", self.obj_type, other.obj_type)

    def __mul__(self, other):
        if other.obj_type == ObjectType.NUM:
            return String_Object(self.value * other.value)
        else:
            raise InternalTypeError("*", self.obj_type, other.obj_type)

    def __truediv__(self, other):
        raise InternalTypeError("/", self.obj_type, other.obj_type)

    def __and__(self, other):
        return Bool_Object(bool(self) and bool(other))

    def __or__(self, other):
        return Bool_Object(bool(self) or bool(other))

    def __neg__(self):
        raise InternalTypeError("-", self.obj_type)

    def __invert__(self):
        return Bool_Object(not bool(self))

    def __bool__(self):
        if self.str_len == 0:
            return False
        return True


class Num_Object(Internal_Object):
    """
    Numeric internal object used to store integers.
    """

    def __init__(self, value):
        self.value = value
        self.obj_type = ObjectType.NUM

    def __repr__(self):
        return "NUM_OBJ {}".format(self.value)

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        assert (
            other is not None
        ), "Other variable is None in __eq__ method for numeric objects."
        return Bool_Object(
            self.obj_type == other.obj_type and self.value == other.value
        )

    def __ne__(self, other):
        return ~(self.__eq__(other))

    def __add__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Num_Object(self.value + other.value)

        raise InternalTypeError("+", self.obj_type, other.obj_type)

    def __sub__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Num_Object(self.value - other.value)

        raise InternalTypeError("-", self.obj_type, other.obj_type)

    def __mul__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Num_Object(self.value * other.value)
        elif other.obj_type in ObjectType.STR:
            return String_Object(other.value * self.value)
        else:
            raise InternalTypeError("*", self.obj_type, other.obj_type)

    def __truediv__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Num_Object(self.value / other.value)

        raise InternalTypeError("/", self.obj_type, other.obj_type)

    def __lt__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Bool_Object(self.value < other.value)

        raise InternalTypeError("<", self.obj_type, other.obj_type)

    def __le__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Bool_Object(self.value <= other.value)
        raise InternalTypeError("<=", self.obj_type, other.obj_type)

    def __gt__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Bool_Object(self.value > other.value)

        raise InternalTypeError(">", self.obj_type, other.obj_type)

    def __ge__(self, other):
        if other.obj_type in [ObjectType.NUM, ObjectType.BOOL]:
            return Bool_Object(self.value >= other.value)
        raise InternalTypeError(">=", self.obj_type, other.obj_type)

    def __neg__(self):
        return Num_Object(-self.value)

    def __invert__(self):
        return Bool_Object(not self.value)

    # These do not override the "and"/"or" keywords but instead override "&" and "|"
    def __and__(self, other):
        return Bool_Object(bool(self) and bool(other))

    def __or__(self, other):
        return Bool_Object(bool(self) or bool(other))

    def __bool__(self):
        return bool(self.value)


class Array_Object(Internal_Object):
    """
    Array internal object used to store a variable amount of elements.
    """

    def __init__(self, elements):
        self.elements = elements
        self.size = len(elements)
        self.obj_type = ObjectType.ARRAY

    def __repr__(self):
        return "ARRAY_OBJ elements: {}, size: {}".format(self.elements, self.size)

    def __str__(self):
        arr_str = "["
        for idx in range(0, self.size - 1):
            arr_str += str(self.elements[idx]) + ", "
        arr_str += str(self.elements[self.size - 1])
        arr_str += "]"
        return arr_str

    def __eq__(self, other):
        assert other is not None, "Other variable in __eq__ function is None."
        if self.obj_type == other.obj_type:
            # Check if size of both arrays are equal
            if self.size == other.size:
                # Deep comparison of each element inside
                for idx in range(0, self.size):
                    if self.elements[idx] != other.elements[idx]:
                        return Bool_Object(False)
                return Bool_Object(True)
            else:
                return Bool_Object(False)
        else:
            return Bool_Object(False)

    # def __ne__(self, other):
    #     assert other is not None, "Other variable in __eq__ function is None."
    #     return ~(self.__eq__(other))

    def __lt__(self, other):
        raise InternalTypeError("<", self.obj_type, other.obj_type)

    def __le__(self, other):
        raise InternalTypeError("<=", self.obj_type, other.obj_type)

    def __ne__(self, other):
        return not (self.__eq__(other))

    def __gt__(self, other):
        raise InternalTypeError(">", self.obj_type, other.obj_type)

    def __ge__(self, other):
        raise InternalTypeError(">=", self.obj_type, other.obj_type)

    def __add__(self, other):
        if other.obj_type == ObjectType.ARRAY:
            self.elements.extend(other.elements)
        else:
            self.elements.append(other)

    def __sub__(self, other):
        raise InternalTypeError("-", self.obj_type, other.obj_type)

    def __mul__(self, other):
        raise InternalTypeError("*", self.obj_type, other.obj_type)

    def __truediv__(self, other):
        raise InternalTypeError("/", self.obj_type, other.obj_type)

    def __and__(self, other):
        return Bool_Object(bool(self) and bool(other))

    def __or__(self, other):
        return Bool_Object(bool(self) or bool(other))

    def __neg__(self):
        raise InternalTypeError(
            "-",
            self.obj_type,
        )

    def __invert__(self):
        return Bool_Object(not self.__bool__())

    def __bool__(self):
        if self.size == 0:
            return False
        return True


class Return_Object(Internal_Object):
    """
    Internal object used to represent return objects within the interpreter.
    """

    def __init__(self, return_val=None):
        self.value = return_val
        self.obj_type = ObjectType.RETURN

    def __str__(self):
        return "RETURN_OBJ {}".format(self.value)

    def __repr__(self):
        return "RETURN_OBJ {}".format(self.value)

    def __eq__(self, other):
        assert (
            other is not None
        ), "Other variable in __eq__ method for a return object is None."
        return Bool_Object(
            self.obj_type == other.obj_type and self.value == other.value
        )


class Break_Object(Internal_Object):
    """
    Internal object used to break statements produced during code execution.
    """

    def __init__(self):
        self.obj_type = ObjectType.BREAK

    def __str__(self):
        return "BREAK_OBJ"

    def __repr__(self):
        return "BREAK_OBJ"

    def __eq__(self, other):
        return self.obj_type == other.obj_type


class Continue_Object(Internal_Object):
    """
    Internal object used to continue statements produced during code execution.
    """

    def __init__(self):
        self.obj_type = ObjectType.CONTINUE

    def __str__(self):
        return "CONTINUE_OBJ"

    def __repr__(self):
        return "CONTINUE_OBJ"

    def __eq__(self, other):
        return self.obj_type == other.obj_type


class Func_Object(Internal_Object):
    """
    Internal object used to represent a function.
    """

    def __init__(self, name, arg_symbols, body, env):
        self.obj_type = ObjectType.FUNC
        self.name = name
        self.args = arg_symbols
        self.arity = len(arg_symbols)
        self.body = body
        # Reference to the current env
        self.env = env

    def __str__(self):
        return "<function object {}>".format(self.name)
        # return "FUNC_OBJ: Name: {}, Args: {}, Arity: {}".format(
        #     self.name, self.args, self.arity,
        # )


class Native_Func_Object(Internal_Object):
    """
    Internal object used to represent a function.
    """

    def __init__(self, func):
        self.obj_type = ObjectType.NATIVE_FUNC
        self.name = func.__name__
        self.arity = func.__code__.co_argcount
        self.body = func

    def __str__(self):
        return "<native function object {}>".format(self.name)


class Class_Def_Object(Internal_Object):
    def __init__(self, class_name, class_methods):
        "setup class object"
        self.obj_type = ObjectType.CLASS_DEF
        self.class_name = class_name
        self.class_methods = class_methods

    def __str__(self):
        return "<class definition object {}>".format(self.class_name)
        # return "CLASS_OBJ: Name: {}".format(self.class_name)


class Module_Object(Internal_Object):
    def __init__(self, module_name, module_path, module_contents, import_as=None):
        "Internal object representing an imported module."
        self.name = module_name
        self.import_as = import_as
        self.path = module_path
        self.namespace = module_contents
        self.obj_type = ObjectType.MODULE

    def __str__(self):
        if self.import_as != self.name:
            return "<module object {} imported as {}>".format(self.name, self.import_as)
        else:
            return "<module object {}>".format(self.name)


class Class_Method_Object(Internal_Object):
    """
    Internal object used to represent a class function.
    """

    def __init__(self, name, arg_symbols, body, class_env):
        self.obj_type = ObjectType.CLASS_METHOD
        self.name = name
        self.args = arg_symbols
        self.arity = len(arg_symbols)
        self.body = body
        # Reference to the class env
        self.class_env = class_env

    def __str__(self):
        return "<class method object {}>".format(self.name)


class Class_Instance_Object(Internal_Object):
    def __init__(self, class_name, class_methods):
        "Object representing a class instance."
        self.obj_type = ObjectType.CLASS_INSTANCE
        self.class_name = class_name
        self.namespace = Scope(None)

        # Register all class methods in the internal environment
        for method in class_methods:
            self.namespace.new_variable(
                method.name,
                Class_Method_Object(
                    method.name, method.args, method.body, self.namespace
                ),
            )

    def __str__(self):
        return "<class instance object {}>".format(self.class_name)

    def get_field(self, field_name):
        return self.namespace.lookup_symbol(field_name)



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

""" Module contains error classes used to transfer control and signal errors
encountered during execution."""


class InternalErr(Exception):
    """
    Base class for all internal errors used by the interpreter.
    """

    def __init__(
        self,
    ):
        raise NotImplementedError()


class InternalTypeError(InternalErr):
    def __init__(self, operation, left_type, right_type=None):
        """Class representing an internal error encountered during and basic operations
        between one or more types.
        """
        self.operation = operation
        self.left = left_type
        self.right = right_type
        self.err_msg = str()

        if self.right is None:
            self.err_msg = "The operation {} is not allowed on a {}!".format(
                self.operation, str(self.left)
            )
        else:
            self.err_msg = (
                "The operation {} is not allowed between a {} and a {}!".format(
                    self.operation, str(self.left), str(self.right)
                )
            )

    def __str__(self):
        return "Typecheck Error: {}".format(self.err_msg)

    def __repr__(self):
        "Internal Runtime Error: Operation {}, Left Side: {}, Right Side: {}".format(
            self.operation, self.left, self.right
        )


class InternalRuntimeErr(InternalErr):
    """
    Class representing an internal error encountered during runtime.
    """

    def __init__(self, message):
        "Class representing internal errors encountered during runtime."
        self.message = message

    def __str__(self):
        return "Runtime Error: {}".format(self.message)

    def __repr__(self):
        "Internal Runtime Error: {}".format(self.message)


class InternalParseErr(InternalErr):
    """
    Class representing an internal error encountered during parsing/lexing stages.
    """

    def __init__(
        self,
        line_num,
        col_num,
        original_text,
        expected_tokens,
        got_token,
    ):
        "Class representing internal errors encountered during parsing/lexing stage."
        self.line_num = line_num
        self.col_num = col_num
        self.source_text_lines = original_text.split("\n")
        self.got_token = got_token
        self.expected_tokens = expected_tokens

        wanted_tokens = ""
        if isinstance(self.expected_tokens, list):
            for tok in self.expected_tokens:
                wanted_tokens += str(tok)
        else:
            wanted_tokens = str(self.expected_tokens)

        self.err_msg = "Expected a '{}' token but received '{}'".format(
            wanted_tokens, self.got_token
        )

    def __str__(self):
        return "Parse Error: Line: {}, Column: {}\n\n  {}\n\nExplanation: {}".format(
            self.line_num,
            self.col_num,
            self.source_text_lines[self.line_num],
            self.err_msg,
        )

    def __repr__(self):
        return (
            "Internal Parse Error: Line: {}, Column: {}\n\n  {}\n\n" "Explanation: {}"
        ).format(
            self.line_num,
            self.col_num,
            self.source_text_lines[self.line_num],
            self.err_msg,
        )


class InternalTokenErr(InternalErr):
    """
    Class representing an internal error encountered during lexing stages.
    """

    def __init__(self, line_num, col_num, original_text, err_details):
        "Class representing internal errors encountered during lexing stage."
        self.line_num = line_num
        self.col_num = col_num
        self.text = original_text.split("\n")
        self.err_details = err_details
        print("error line", self.text[self.line_num])

    def __str__(self):
        return "Token Error: Error on line {}, column {}\n  {}\n{}\n{}".format(
            self.line_num,
            self.col_num,
            self.text[self.line_num],
            "      _".rjust(self.col_num - 1, " "),
            self.err_details,
        )

    def __repr__(self):
        return "Internal Token Error: Error on line {}, column {}\n  {}".format(
            self.line_num, self.col_num, self.text[self.line_num]
        )



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

"""
Module containing the parser class used to produce the AST to be evaluated
by the interpreter.
"""

from zai.ast_nodes import (
    This_Node,
    String_Node,
    Num_Node,
    Array_Node,
    Array_Access_Node,
    Nil_Node,
    Bool_Node,
    Arith_Bin_Node,
    ID_Node,
    Dot_Bin_Node,
    Bracket_Node,
    Unary_Node,
    Incr_Node,
    Decr_Node,
    Call_Node,
    AddAssign_Node,
    SubAssign_Node,
    New_Assign_Bin_Node,
    If_Node,
    Print_Node,
    Func_Node,
    Scope_Block_Node,
    While_Node,
    Break_Node,
    Continue_Node,
    Do_While_Node,
    Import_Node,
    Program_Node,
    Switch_Node,
    Class_Def_Node,
    Class_Method_Node,
    Return_Node,
    Replace_Assign_Bin_Node,
    Relop_Bin_Node,
    Eq_Bin_Node,
    Logic_Bin_Node,
)
from zai.tokens import Tok_Type
from zai.internal_error import InternalParseErr


class Parser:
    """
    Parse and produce an AST from the provided token stream.
    """

    def __init__(self, tokens, original_text):
        self.original_text = original_text
        self.tokens = tokens
        self.tok_len = len(tokens)
        self.curr_idx = 0
        self.curr_tok = self.tokens[self.curr_idx]

        self.ast = None

    def peek(self, n=1):
        """
        Look ahead and return the Nth token in the token stream. If a numeric
        argument is not provided, return the next token in the stream. If
        there is no next token, return None.
        """
        if self.curr_idx + n < self.tok_len:
            return self.tokens[self.curr_idx + n]
        else:
            return None

    def advance(self, n=1):
        """
        Advance the current token being used by N places. If a numeric
        argument is not provided, return the next token in the stream.
        If there is no next token, return None."""
        if self.curr_idx + n < self.tok_len:
            self.curr_idx += n
            self.curr_tok = self.tokens[self.curr_idx]
            return self.curr_tok
        else:
            return None

    def match(self, *args):
        """
        Check if the type of the current token being used matches with the provided
        token types. The token types to be checked can be either in a  list or a single
        token.

        There is no error produced from type mismatched yet."""
        if self.curr_tok in args:
            curr_token = self.curr_tok
            self.advance(1)
            return curr_token
        else:
            raise InternalParseErr(
                self.curr_tok.line_num,
                self.curr_tok.col_num,
                self.original_text,
                list(args),
                self.curr_tok.tok_type,
            )

    def atom(self):
        """
        Parse an atom(pritive) object.
        Grammar:
        atom := | NUM | STR | BOOL | ARRAY | "nil"
        ;; Primitives
        BOOL := "false" | "true"
        ID := LETTER (LETTER | NUM | "-" | "$" | "@")*
        NUM := DIGIT (DIGIT)*
        DIGIT := 1 | 2 | 3| 4 | 5 | 6 | 7 | 8 | 9 | 0
        STR := "\"" (LETTER | NUM )* "\""
        ARRAY := "[" ( or_expr ( "," or_expr )* )? "]"
        """
        if self.curr_tok.tok_type == Tok_Type.THIS:
            node = self.match(Tok_Type.THIS)
            return This_Node()
        elif self.curr_tok.tok_type == Tok_Type.DQUOTE:
            self.match(Tok_Type.DQUOTE)
            str_token = self.match(Tok_Type.STRING)
            node = String_Node(str_token.lexeme)
            self.match(Tok_Type.DQUOTE)
            return node
        elif self.curr_tok.tok_type == Tok_Type.NUM:
            node = self.match(Tok_Type.NUM)
            return Num_Node(node.lexeme)
        elif self.curr_tok.tok_type == Tok_Type.LSQUARE:
            self.match(Tok_Type.LSQUARE)
            array_elem = list()
            while self.curr_tok.tok_type != Tok_Type.RSQUARE:
                arg_value = self.or_expr()
                array_elem.append(arg_value)
                if self.curr_tok.tok_type != Tok_Type.RSQUARE:
                    self.match(Tok_Type.COMMA)
            self.match(Tok_Type.RSQUARE)
            return Array_Node(array_elem)
        elif self.curr_tok.tok_type == Tok_Type.NIL:
            self.match(Tok_Type.NIL)
            return Nil_Node()
        else:
            node = self.match(Tok_Type.TRUE, Tok_Type.FALSE)
            return Bool_Node(node.tok_type)

    def arglist(self):
        """
        Parse an arglist rule.
        Grammar:
        arglist := (or_expr ("," or_expr)*)?
        """
        func_args = list()
        while self.curr_tok.tok_type != Tok_Type.RROUND:
            arg_value = self.or_expr()
            func_args.append(arg_value)
            if self.curr_tok.tok_type != Tok_Type.RROUND:
                self.match(Tok_Type.COMMA)
        return func_args

    def access_ident(self):
        # Create ID node
        node = self.match(Tok_Type.ID)
        left = ID_Node(node.lexeme)

        while self.curr_tok.tok_type in [
            Tok_Type.DOT,
            Tok_Type.LSQUARE,
        ]:
            if self.curr_tok.tok_type == Tok_Type.DOT:
                self.match(Tok_Type.DOT)
                property_name = ID_Node(self.match(Tok_Type.ID).lexeme)
                left = Dot_Bin_Node(left, property_name)
            elif self.curr_tok.tok_type == Tok_Type.LSQUARE:
                self.match(Tok_Type.LSQUARE)
                arr_idx = self.atom()
                self.match(Tok_Type.RSQUARE)
                left = Array_Access_Node(left, arr_idx)

        return left

    def access_this(self):
        self.match(Tok_Type.THIS)
        left = ID_Node("this")

        while self.curr_tok.tok_type in [
            Tok_Type.DOT,
            Tok_Type.LSQUARE,
        ]:
            if self.curr_tok.tok_type == Tok_Type.DOT:
                self.match(Tok_Type.DOT)
                property_name = ID_Node(self.match(Tok_Type.ID).lexeme)
                left = Dot_Bin_Node(left, property_name)
            elif self.curr_tok.tok_type == Tok_Type.LSQUARE:
                self.match(Tok_Type.LSQUARE)
                arr_idx = self.atom()
                self.match(Tok_Type.RSQUARE)
                left = Array_Access_Node(left, arr_idx)

        return left

    def access(self):
        if self.curr_tok.tok_type == Tok_Type.ID:
            return self.access_ident()
        elif self.curr_tok.tok_type == Tok_Type.THIS:
            return self.access_this()

    def call_or_access(self):
        """
        Parse a call or access rule.
        Grammar:
        call_or_access := ( ID | "this" ) ( "(" arglist ")" | "." ID )*
        """

        access_node = self.access()
        if self.curr_tok.tok_type == Tok_Type.LROUND:
            self.match(Tok_Type.LROUND)
            func_args = self.arglist()
            self.match(Tok_Type.RROUND)
            return Call_Node(access_node, func_args)
        else:
            return access_node

    def factor(self):
        """
        Parse a factor rule.
        Grammar:

        factor := "(" expr ")" | unary_op factor | atom
        unary_op := "!" "-"
        """
        # TODO: There will be a problem with the first if statement. Nonetype does not
        # does not have a "tok_type" so it will fail if there is no next token.
        if self.curr_tok.tok_type in [
            Tok_Type.ID,
            Tok_Type.THIS,
        ]:
            return self.call_or_access()
        elif self.curr_tok.tok_type == Tok_Type.LROUND:
            self.match(Tok_Type.LROUND)
            expr = self.or_expr()
            self.match(Tok_Type.RROUND)
            return Bracket_Node(expr)
        elif self.curr_tok.tok_type in [Tok_Type.BANG, Tok_Type.MINUS]:
            op = self.match(Tok_Type.BANG, Tok_Type.MINUS)
            fact = self.factor()
            return Unary_Node(op.tok_type, fact)
        # TODO: Handle case of it being an "ID" token instead of just num.
        elif self.curr_tok.tok_type in [Tok_Type.NUM, Tok_Type.ID] and self.peek() in [
            Tok_Type.INCR,
            Tok_Type.DECR,
        ]:
            # Evaluate the initial number token
            node = self.atom()

            # "++" or "--" operators can be nested
            while self.curr_tok.tok_type in [Tok_Type.INCR, Tok_Type.DECR]:
                op = self.match(Tok_Type.INCR, Tok_Type.DECR)
                if op.tok_type == Tok_Type.INCR:
                    node = Incr_Node(node)
                elif op.tok_type == Tok_Type.DECR:
                    node = Decr_Node(node)
            return node
        else:
            return self.atom()

    def term(self):
        """
        Parse a term rule.
        Grammar:
        term := factor (("*" | "/") factor)*
        """
        left = self.factor()
        while self.curr_tok.tok_type in [Tok_Type.MUL, Tok_Type.DIV]:
            op = self.match(Tok_Type.MUL, Tok_Type.DIV)
            right = self.term()
            left = Arith_Bin_Node(left, op.tok_type, right)

        return left

    def add_expr(self):
        """
        Parse an addition expression:
        Grammar:
        add_expr := term (("+" | "-") term)*
        """
        left = self.term()
        while self.curr_tok.tok_type in [Tok_Type.PLUS, Tok_Type.MINUS]:
            op = self.match(Tok_Type.PLUS, Tok_Type.MINUS)
            right = self.term()
            left = Arith_Bin_Node(left, op.tok_type, right)

        return left

    def rel_expr(self):
        """
        Parse a relative expression:
        Grammar:
        rel_expr := add_expr ( ("<" | ">" | "<=" | ">=") add_expr)*
        """

        left = self.add_expr()
        while self.curr_tok.tok_type in [
            Tok_Type.GT,
            Tok_Type.GTE,
            Tok_Type.LT,
            Tok_Type.LTE,
        ]:
            op = self.match(Tok_Type.GT, Tok_Type.GTE, Tok_Type.LT, Tok_Type.LTE)
            right = self.add_expr()
            left = Relop_Bin_Node(left, op.tok_type, right)

        return left

    def eq_expr(self):
        """
        Parse an equality expression:
        Grammar:
        eq_expr := rel_expr ( ("=="|"!=") rel_expr)*
        """
        left = self.rel_expr()
        while self.curr_tok.tok_type in [Tok_Type.EQ, Tok_Type.NEQ]:
            op = self.match(Tok_Type.EQ, Tok_Type.NEQ)
            right = self.rel_expr()
            left = Eq_Bin_Node(left, op.tok_type, right)

        return left

    def and_expr(self):
        """
        Parse an equality expression:
        Grammar:
        and_expr := eq_expr ("&&" eq_expr)*
        """
        left = self.eq_expr()
        while self.curr_tok.tok_type == Tok_Type.AND:
            op = self.match((Tok_Type.AND))
            right = self.eq_expr()
            left = Logic_Bin_Node(left, op.tok_type, right)

        return left

    def or_expr(self):
        """
        Parse an equality expression:
        Grammar:
        or_expr := and_expr ("||" and_expr)*
        """
        left = self.and_expr()
        while self.curr_tok.tok_type == Tok_Type.OR:
            op = self.match(Tok_Type.OR)
            right = self.and_expr()
            left = Logic_Bin_Node(left, op.tok_type, right)

        return left

    def expr_statement(self):
        node = self.or_expr()
        self.match(Tok_Type.SEMIC)
        return node

    def reassign_or_eval_statement(self):
        # There are a couple of cases to handle here:
        # 1. This is a regular expression like "a + 4;"
        # 2. This is a simple variable access like "a.b;" or "a[3];"
        # 3. This is a reassign operation like "a.b = 4;"
        var_name = self.or_expr()
        if isinstance(var_name, (Dot_Bin_Node, ID_Node, This_Node, Array_Access_Node)):
            # TODO: Throw error if we get a "This" node without anything else attached
            # to it.
            if self.curr_tok.tok_type == Tok_Type.ASSIGN:
                self.match(Tok_Type.ASSIGN)
                value = self.expr_statement()
                if isinstance(var_name, (ID_Node, Array_Access_Node)):
                    a = Replace_Assign_Bin_Node(None, var_name, value)
                    return a
                else:
                    a = Replace_Assign_Bin_Node(var_name.left, var_name.right, value)
                    return a
            # TODO: Add augmented assign cases here.
            elif self.curr_tok.tok_type == Tok_Type.ADDASSIGN:
                self.match(Tok_Type.ADDASSIGN)
                value = self.expr_statement()
                if isinstance(var_name, ID_Node):
                    return AddAssign_Node(None, var_name, value)
                else:
                    # Decompose the nodes
                    return AddAssign_Node(var_name.left, var_name.right, value)
            elif self.curr_tok.tok_type == Tok_Type.SUBASSIGN:
                self.match(Tok_Type.SUBASSIGN)
                value = self.expr_statement()
                if isinstance(var_name, ID_Node):
                    return SubAssign_Node(None, var_name, value)
                else:
                    # Decompose the nodes
                    return SubAssign_Node(var_name.left, var_name.right, value)
            else:
                # Case of the node not being an assignment node but a variable access
                self.match(Tok_Type.SEMIC)
                return var_name
        else:
            # Case of a regular expression like "a + 4;"
            self.match(Tok_Type.SEMIC)
            return var_name

    def new_asssign_statement(self):
        self.match(Tok_Type.LET)
        symbol_path = self.call_or_access()
        self.match(Tok_Type.ASSIGN)
        value = self.expr_statement()
        # Check if it is a single node
        if isinstance(symbol_path, ID_Node):
            return New_Assign_Bin_Node(None, symbol_path, value)
        else:
            # Decompose the nodes
            return New_Assign_Bin_Node(symbol_path.left, symbol_path.right, value)

    def if_statement(self):
        """
        Parse an if statement.
        Grammar:

        if_stmnt := "if" "(" expression ")" statement ( "else" statement )?
        """
        # List used to store all "if condition then" or "elif condition then" pairs.
        conditions = list()

        # Parse "If ... then ..."
        self.match(Tok_Type.IF)
        self.match(Tok_Type.LROUND)
        if_cond = self.or_expr()
        self.match(Tok_Type.RROUND)
        then_block = self.block()
        conditions.append({"cond": if_cond, "block": then_block})

        # Parse "elif ... then ..."
        while self.curr_tok.tok_type == Tok_Type.ELIF:
            self.match(Tok_Type.ELIF)
            self.match(Tok_Type.LROUND)
            cond = self.or_expr()
            self.match(Tok_Type.RROUND)
            then_block = self.block()
            conditions.append({"cond": cond, "block": then_block})

        else_block = None
        if self.curr_tok.tok_type == Tok_Type.ELSE:
            self.match(Tok_Type.ELSE)
            else_block = self.block()

        node = If_Node(conditions, else_block)
        return node

    def print_statement(self):
        """
        Parse a print statement.
        Grammar:
        """

        self.match(Tok_Type.PRINT)
        expr = self.expr_statement()
        return Print_Node(expr)

    # TODO: Refactor match function and finish parsing functions
    def func_def(self):
        """
        Parse a function definition.
        Grammar:

        func_def := "func" ID "(" ID? ("," ID)* ")" "{" statement* "}"
        """

        self.match(Tok_Type.FUNC)
        func_name = self.match(Tok_Type.ID).lexeme
        self.match(Tok_Type.LROUND)

        # Collect argument symbols for the function
        arg_symbols = list()
        if self.curr_tok.tok_type == Tok_Type.ID:
            func_arg = self.match(Tok_Type.ID)
            arg_symbols.append(func_arg)
            if self.curr_tok.tok_type == Tok_Type.COMMA:
                while self.curr_tok.tok_type != Tok_Type.RROUND:
                    self.match(Tok_Type.COMMA)
                    func_arg = self.match(Tok_Type.ID)
                    arg_symbols.append(func_arg)

        self.match(Tok_Type.RROUND)
        self.match(Tok_Type.LCURLY)

        func_body = list()
        while self.curr_tok.tok_type != Tok_Type.RCURLY:
            func_stmnt = self.statement()
            func_body.append(func_stmnt)
        self.match(Tok_Type.RCURLY)

        return Func_Node(func_name, arg_symbols, func_body)

    def block(self):
        """
        Parse a block which acts as a nested scope within the program.
        Grammar:

        block := "{" statement* "}"
        """

        self.match(Tok_Type.LCURLY)
        block_stmnts = list()
        while self.curr_tok.tok_type != Tok_Type.RCURLY:
            stmnt = self.statement()
            block_stmnts.append(stmnt)
        self.match(Tok_Type.RCURLY)
        return Scope_Block_Node(block_stmnts)

    def while_statement(self):
        """
        Parse a while statement.
        Grammar:
        while_stmnt := "while" "(" or_expr ")" block_stmnt
        """
        self.match(Tok_Type.WHILE)
        self.match(Tok_Type.LROUND)
        condition = self.or_expr()
        self.match(Tok_Type.RROUND)

        body = self.block()
        return While_Node(condition, body)

    def switch_stmnt_case(self):
        """
        Parse a single switch statement case.
        """
        stmnt_block = list()
        while self.curr_tok.tok_type not in [
            Tok_Type.CASE,
            Tok_Type.DEFAULT,
            Tok_Type.RCURLY,
        ]:
            stmnt = self.statement()
            stmnt_block.append(stmnt)

        # We can reuse the block node to evaluate the statements inside of a switch
        # case. The only difference between the two is how parsing is done. Block nodes
        # require curly brackets while switch case statements do not necessarily
        # require them.
        return Scope_Block_Node(stmnt_block)

    def switch_statement(self):
        """
        Parse a switch statement rule.
        Grammar:
        switch_stmnt := "switch" "(" or_expr ")"
                        ("case"  or_expr ":" block_stmnt)*
                        "default" ":" block_stmnt
        """
        self.match(Tok_Type.SWITCH)
        self.match(Tok_Type.LROUND)
        switch_cond = self.or_expr()
        self.match(Tok_Type.RROUND)
        self.match(Tok_Type.LCURLY)

        cases = list()
        while self.curr_tok.tok_type == Tok_Type.CASE:
            self.match(Tok_Type.CASE)
            test_cond = self.or_expr()
            self.match(Tok_Type.COLON)
            block = self.switch_stmnt_case()
            cases.append(tuple((test_cond, block)))

        self.match(Tok_Type.DEFAULT)
        self.match(Tok_Type.COLON)
        default_block = self.switch_stmnt_case()
        self.match(Tok_Type.RCURLY)

        return Switch_Node(switch_cond, cases, default_block)

    def class_def(self):
        """
        Parse a class definition rule.
        Grammar:
        class_def := "class" ID "{" func_def* "}"
        """
        self.match(Tok_Type.CLASS)
        class_name = self.match(Tok_Type.ID).lexeme
        self.match(Tok_Type.LCURLY)

        class_methods = list()
        while self.curr_tok.tok_type != Tok_Type.RCURLY:
            func_node = self.func_def()
            # Cast regular function node to method node.
            # The parsing procedure is the same for both so there is no point
            # in rewriting the code in a separate parsing procedure.
            method_node = Class_Method_Node(
                func_node.name, func_node.args, func_node.body
            )
            class_methods.append(method_node)

        self.match(Tok_Type.RCURLY)

        return Class_Def_Node(class_name, class_methods)

    def flow_statement(self):
        """
        Parse a flow statement rule.
        Grammar:
        flow_stmnt := ("break" | "continue" | "return" ( or_expr )? ) ";"
        """
        node = None
        if self.curr_tok.tok_type == Tok_Type.RETURN:
            return_val = None
            self.match(Tok_Type.RETURN)
            # Check if the return statement contains an expression or not.
            if self.curr_tok.tok_type != Tok_Type.SEMIC:
                return_val = self.or_expr()

            node = Return_Node(return_val)
        elif self.curr_tok.tok_type == Tok_Type.BREAK:
            self.match(Tok_Type.BREAK)
            node = Break_Node()
        elif self.curr_tok.tok_type == Tok_Type.CONTINUE:
            self.match(Tok_Type.CONTINUE)
            node = Continue_Node()

        self.match(Tok_Type.SEMIC)
        return node

    def do_while_statement(self):
        """
        Parse a do-while statement.
        Grammar:
        do_while_stmnt := "do" block_stmnt "while" "(" or_expr ")"
        """
        self.match(Tok_Type.DO)
        body = self.block()
        self.match(Tok_Type.WHILE)
        self.match(Tok_Type.LROUND)
        test_cond = self.or_expr()
        self.match(Tok_Type.RROUND)
        self.match(Tok_Type.SEMIC)
        return Do_While_Node(test_cond, body)

    def import_statement(self):
        """
        Parse an import statement.
        Grammar:
        import_stmnt := "import" ID ("as" ID)?
        """
        self.match(Tok_Type.IMPORT)
        module_name = self.match(Tok_Type.ID).lexeme
        import_name = None
        if self.curr_tok.tok_type == Tok_Type.AS:
            self.match(Tok_Type.AS)
            import_name = self.match(Tok_Type.ID).lexeme

        return Import_Node(module_name, import_name)

    def statement(self):
        """
         Parse a single statement.
         Grammar:
        statement := if_stmnt        |
                     while_stmnt     |
                     for_stmnt       |
                     class_def       |
                     func_def        |
                     block_stmnt     |
                     print_stmnt     |
                     switch_stmnt    |
                     flow_stmnt      |
                     do_while_stmnt  |
                     import_stmnt    |
                     simple_expr
        """
        if self.curr_tok.tok_type == Tok_Type.IF:
            return self.if_statement()
        elif self.curr_tok.tok_type == Tok_Type.FUNC:
            return self.func_def()
        elif self.curr_tok.tok_type == Tok_Type.CLASS:
            return self.class_def()
        elif self.curr_tok.tok_type == Tok_Type.WHILE:
            return self.while_statement()
        elif self.curr_tok.tok_type == Tok_Type.SWITCH:
            return self.switch_statement()
        elif self.curr_tok.tok_type == Tok_Type.LCURLY:
            return self.block()
        elif self.curr_tok.tok_type == Tok_Type.PRINT:
            return self.print_statement()
        elif self.curr_tok.tok_type == Tok_Type.SWITCH:
            return self.switch_statement()
        elif self.curr_tok.tok_type == Tok_Type.DO:
            return self.do_while_statement()
        elif self.curr_tok.tok_type == Tok_Type.IMPORT:
            return self.import_statement()
        elif self.curr_tok.tok_type in [
            Tok_Type.RETURN,
            Tok_Type.CONTINUE,
            Tok_Type.BREAK,
        ]:
            return self.flow_statement()
        elif self.curr_tok.tok_type == Tok_Type.LET:
            return self.new_asssign_statement()
        elif self.curr_tok.tok_type in [Tok_Type.ID, Tok_Type.THIS]:
            return self.reassign_or_eval_statement()
        else:
            return self.expr_statement()

    def program(self):
        """
        Parse a program rule.
        Grammar:
        program := statement*
        """
        prog_node = Program_Node()
        while self.curr_tok.tok_type != Tok_Type.EOF:
            stmnt = self.statement()
            prog_node.add_stmnt(stmnt)
        return prog_node

    def parse(self):
        """
        Parse the provided stream of tokens and return the AST produced.
        """
        ast_root = self.program()
        self.ast = ast_root
        return self.ast



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

"""
Module containing classes related to managing the interpreter environment
and nested scopes.
"""


class Scope:
    def __init__(self, parent):
        self.scope = dict()
        self.parent = parent

    def new_variable(self, var_name, value):
        """
        Instantiate a new variable within the current scope of the environment.
        Return True to indicate success.
        """
        assert (
            var_name is not None
        ), "Variable name of new environment variable is None."
        assert value is not None, "Variable value of new environment variable is None."
        self.scope[var_name] = value
        return True

    def replace_variable(self, var_name, value):
        """
        Replace the value of an existing variable within the environment.
        Return True or False if the value exists/does not exist within
        the current environment.
        """
        assert var_name is not None, "Variable name to be replaced is None."
        assert value is not None, "Variable value to be replaced is None"
        if var_name in self.scope.keys():
            self.scope[var_name] = value
            return True
        elif self.parent is None:
            return False
        else:
            return self.parent.replace_variable(var_name, value)

    def lookup_symbol(self, symbol):
        """
        Lookupt the value of a symbol in the current scope and all parent scopes
        and return it. If the symbol does not exist, return None.
        """
        value = self.scope.get(symbol, None)
        if value is None and self.parent is not None:
            return self.parent.lookup_symbol(symbol)
        return value

    def merge_scopes(self, new_scope):
        """
        Merge two scope objects into one.
        """
        for k, v in new_scope.scope.items():
            self.scope[k] = v

    def __str__(self):
        return str(self.scope)


class Environment:
    """
    Class responsible for managing a stack of environment scopes.
    """

    def __init__(self):
        self.scopes = list()
        # Add the global scope
        self.scopes.append(Scope(None))
        self.stack_height = 0

    def peek(self):
        """
        Return the scope on the top of the scope stack.
        """
        if self.stack_height > -1:
            return self.scopes[self.stack_height]

    def exit_scope(self):
        """
        Remove the most current scope from the scope stack.
        """
        if self.stack_height > -1:
            self.scopes.pop()
            self.stack_height -= 1

    def enter_scope(self, parent_scope=None):
        """
        Put a new scope on top of the scope stack.
        """
        self.scopes.append(Scope(parent_scope))
        self.stack_height += 1

    def __str__(self):
        env_stack = str()
        for scope in self.scopes:
            env_stack += str(scope) + "\n"
        return env_stack



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

import argparse

from zai.vm import YAPL_VM
from pathlib import Path
from sys import exit


def main():
    vm = YAPL_VM()
    arg_parser = argparse.ArgumentParser(
        "yapl",
        description="An interpreter for yapl.",
        add_help="Interpreter for yapl",
    )
    arg_parser.add_argument(
        "file_path",
        nargs="?",
        help="File path leading to file to be interpreted.",
        default=None,
    )
    arg_parser.add_argument(
        "-e",
        "--eval_string",
        nargs=1,
        help="Evaluate a string and exit.",
        required=False,
        type=str,
    )

    args = arg_parser.parse_args()
    if args.eval_string is not None:
        vm.run_string(args.eval_string[0])
        exit(0)
    elif args.file_path is None:
        vm.run_repl()
        exit(0)
    else:
        f_path = Path(args.file_path)
        if f_path.exists() and f_path.is_file():
            file_text = f_path.open().read()
            # print(file_text)
            vm.run_string(file_text)
            exit(0)

        print("ERROR: path {} does not exist or is not a file.".format(args.file_path))
        exit(1)


if __name__ == "__main__":
    main()



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

from enum import Enum, auto


class Tok_Type(Enum):
    """ Class representing all token types generated by the lexer. """

    # Atom types
    STRING = auto()
    ID = auto()
    BOOL = auto()
    NUM = auto()

    COMMA = auto()
    # A semicolon ";"
    SEMIC = auto()
    # A colon ":"
    COLON = auto()
    # Single quote '
    QUOTE = auto()
    # Double quote "
    DQUOTE = auto()
    LROUND = auto()
    RROUND = auto()
    LSQUARE = auto()
    RSQUARE = auto()
    LCURLY = auto()
    RCURLY = auto()
    # Comparison op
    EQ = auto()
    NEQ = auto
    LT = auto()
    GT = auto()
    LTE = auto()
    GTE = auto()

    # arith op
    PLUS = auto()
    MINUS = auto()
    DIV = auto()
    MUL = auto()
    INCR = auto()
    DECR = auto()

    # Boolean op
    BANG = auto()
    TRUE = auto()
    FALSE = auto()
    AND = auto()
    OR = auto()

    # Control flow
    IF = auto()
    ELSE = auto()
    ELIF = auto()
    WHILE = auto()
    FOR = auto()
    SWITCH = auto()
    CASE = auto()
    DO = auto()
    DEFAULT = auto()
    BREAK = auto()
    CONTINUE = auto()

    # Variable Assignment
    ASSIGN = auto()
    ADDASSIGN = auto()
    SUBASSIGN = auto()
    LET = auto()

    # MISC.
    FUNC = auto()
    CLASS = auto()
    DOT = auto()
    THIS = auto()
    PRINT = auto()
    RETURN = auto()
    NIL = auto()
    IMPORT = auto()
    AS = auto()
    EOF = auto()

    def __str__(self):
        tok_type_to_str = {
            "STRING": "string",
            "ID": "identifier",
            "BOOL": "boolean",
            "NUM": "number",
            "COMMA": ",",
            "SEMIC": ";",
            "COLON": ":",
            "QUOTE": "'",
            "DQUOTE": '"',
            "LROUND": "(",
            "RROUND": ")",
            "LSQUARE": "[",
            "RSQUARE": "]",
            "LCURLY": "{",
            "RCURLY": "}",
            "EQ": "==",
            "NEQ": "!=",
            "LT": "<",
            "GT": ">",
            "LTE": "<=",
            "GTE": ">",
            "PLUS": "+",
            "MINUS": "-",
            "DIV": "/",
            "MUL": "*",
            "INCR": "++",
            "DECR": "--",
            "BANG": "!",
            "TRUE": "true",
            "FALSE": "false",
            "AND": "&&",
            "OR": "||",
            "IF": "if keyword",
            "ELSE": "else keyword",
            "ELIF": "elif keyword",
            "WHILE": "while keyword",
            "FOR": "for keyword",
            "SWITCH": "switch keyword",
            "CASE": "case keyword",
            "DO": "do keyword",
            "DEFAULT": "default keyword",
            "BREAK": "break keyword",
            "CONTINUE": "continue keyword",
            "ASSIGN": "=",
            "ADDASSIGN": "+=",
            "SUBASSIGN": "-=",
            "LET": "let keyword",
            "FUNC": "func keyword",
            "CLASS": "class keyword",
            "DOT": ".",
            "THIS": "this keyword",
            "PRINT": "print keyword",
            "RETURN": "return keyword",
            "NIL": "nil",
            "IMPORT": "import keyword",
            "AS": "as keyword",
            "EOF": "End Of File Marker",
        }
        return tok_type_to_str[self.name]


class Token:
    """Token class used for all tokens generated by the lexer."""

    def __init__(self, tok_type, lexeme=None, line_num=None, col_num=None):
        """Generate a new token.

        Args:
            tok_type : The type of token.
            lexeme : Lexeme stored by the token. Defaults to None.
            line_num : Line number where lexeme was encountered.. Defaults to None.
            col_num : Column number where lexeme starts. Defaults to None.
        """
        self.tok_type = tok_type
        self.lexeme = lexeme
        self.line_num = line_num
        self.col_num = col_num

    def __str__(self):
        if self.lexeme is not None:
            return "(T:{} Le:{} C:{} L:{})".format(
                self.tok_type, self.lexeme, self.col_num, self.line_num
            )
        else:
            return "(T:{} C:{} L:{})".format(self.tok_type, self.col_num, self.line_num)

    def __repr__(self):
        if self.lexeme is not None:
            return "(T:{} Le:{} C:{} L:{})".format(
                self.tok_type, self.lexeme, self.col_num, self.line_num
            )
        else:
            return "(T:{} C:{} L:{})".format(self.tok_type, self.col_num, self.line_num)

    def __eq__(self, right):
        # Necessary to compare Tok_Type with a token.
        if isinstance(right, Enum):
            if right == self.tok_type:
                return True
            else:
                return False
        elif self.tok_type == right.tok_type and self.lexeme == right.lexeme:
            return True

        return False


# Dictionary containing words reserved by language and their corresponding tokens
keywords = {
    "if": Tok_Type.IF,
    "else": Tok_Type.ELSE,
    "elif": Tok_Type.ELIF,
    "while": Tok_Type.WHILE,
    "for": Tok_Type.FOR,
    "print": Tok_Type.PRINT,
    "true": Tok_Type.TRUE,
    "false": Tok_Type.FALSE,
    "let": Tok_Type.LET,
    "func": Tok_Type.FUNC,
    "and": Tok_Type.AND,
    "or": Tok_Type.OR,
    "switch": Tok_Type.SWITCH,
    "case": Tok_Type.CASE,
    "default": Tok_Type.DEFAULT,
    "class": Tok_Type.CLASS,
    "this": Tok_Type.THIS,
    "return": Tok_Type.RETURN,
    "break": Tok_Type.BREAK,
    "continue": Tok_Type.CONTINUE,
    "do": Tok_Type.DO,
    "nil": Tok_Type.NIL,
    "import": Tok_Type.IMPORT,
    "as": Tok_Type.AS,
}



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

import zai.objects


def object_type(internal_object):
    """
    Return the type of an internal object as a string. If the type is
    unknown, return 'nil'.
    """
    if internal_object.obj_type == zai.objects.ObjectType.ARRAY:
        return zai.objects.String_Object("array")
    elif internal_object.obj_type == zai.objects.ObjectType.NUM:
        return zai.objects.String_Object("number")
    elif internal_object.obj_type == zai.objects.ObjectType.STR:
        return zai.objects.String_Object("string")
    elif internal_object.obj_type == zai.objects.ObjectType.NIL:
        return zai.objects.String_Object("nil")
    elif internal_object.obj_type == zai.objects.ObjectType.MODULE:
        return zai.objects.String_Object("module")
    elif internal_object.obj_type == zai.objects.ObjectType.BOOL:
        return zai.objects.String_Object("boolean")
    elif internal_object.obj_type == zai.objects.ObjectType.FUNC:
        return zai.objects.String_Object("function")
    elif internal_object.obj_type == zai.objects.ObjectType.CLASS_DEF:
        return zai.objects.String_Object("class_def")
    elif internal_object.obj_type == zai.objects.ObjectType.CLASS_INSTANCE:
        return zai.objects.String_Object("class_instance")
    elif internal_object.obj_type == zai.objects.ObjectType.CLASS_METHOD:
        return zai.objects.String_Object("class_method")
    else:
        return zai.objects.Nil_Object()


def str_len(internal_object):
    """
    Determine the length of a string. If the object passed to this function is
    not a string object, return nil.
    """
    if internal_object.obj_type == zai.objects.ObjectType.STR:
        return zai.objects.Num_Object(internal_object.str_len)
    else:
        return zai.objects.Nil_Object()


def mod(operand1, operand2):
    """
    Find the modulus of operand1 and operand2. Return nil if arguments
    are not numbers.
    """
    if (
        operand1.obj_type == zai.objects.ObjectType.NUM
        and operand2.obj_type == zai.objects.ObjectType.NUM
    ):
        if operand1.value == 0 or operand2.value == 0:
            return zai.objects.Nil_Object()
        else:
            return zai.objects.Num_Object(operand1.value % operand2.value)
    else:
        return zai.objects.Nil_Object()


def power(base, exponent):
    """
    Return the argument base raised to the power represented by exponent.
    """
    if (
        base.obj_type == zai.objects.ObjectType.NUM
        and exponent.obj_type == zai.objects.ObjectType.NUM
    ):
        return zai.objects.Num_Object(pow(base.value, exponent.value))
    else:
        return zai.objects.Nil_Object()


def register_functions():
    """
    Transform all native function defined within this module into internal objects
    which can be called by the interpreter.
    """
    registered_functions = list()
    registered_functions.append(zai.objects.Native_Func_Object(str_len))
    registered_functions.append(zai.objects.Native_Func_Object(object_type))
    registered_functions.append(zai.objects.Native_Func_Object(power))
    registered_functions.append(zai.objects.Native_Func_Object(mod))

    return registered_functions



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

"""Module contains a visitor class implementation used to execute the AST
produced by the parser."""
from zai.ast_nodes import Array_Access_Node
from zai.tokens import Tok_Type
from zai.internal_error import InternalRuntimeErr
from zai.env import Environment, Scope
from zai.lexer import Lexer
from zai.parse import Parser
from zai.utils import is_truthy, read_module_contents
from zai.objects import (
    ObjectType,
    Bool_Object,
    Nil_Object,
    Num_Object,
    Func_Object,
    String_Object,
    Return_Object,
    Class_Def_Object,
    Class_Instance_Object,
    Continue_Object,
    Break_Object,
    Array_Object,
    Module_Object,
)


class Visitor:
    def __init__(self, environment):
        """
        Class implementing the visitor pattern which is used to evaluate
        language structures.
        """
        self.env = environment

    def visit(self, ast_root):
        """
        Main entry point for all AST roots.
        """
        return ast_root.accept(self)

    def visit_program(self, node):
        for stmnt in node.stmnts:
            ret_val = stmnt.accept(self)

            if ret_val is not None and ret_val.obj_type in [
                ObjectType.RETURN,
                ObjectType.BREAK,
                ObjectType.CONTINUE,
            ]:
                if ret_val.obj_type == ObjectType.RETURN:
                    msg = (
                        '"return" statement not used outside of a function or class'
                        "method!"
                    )
                    raise InternalRuntimeErr(msg)
                elif ret_val.obj_type == ObjectType.BREAK:
                    msg = '"break" statement not used within a loop or a switch block!'
                    raise InternalRuntimeErr(msg)
                elif ret_val.obj_type == ObjectType.CONTINUE:
                    msg = '"continue" statement not used within a loop!'
                    raise InternalRuntimeErr(msg)
            # else:
            #     return ret_val

    def visit_num(self, node):
        return Num_Object(node.val)

    def visit_symbol(self, node):
        # Retrieve symbol from env
        curr_scope = self.env.peek()
        symbol_val = curr_scope.lookup_symbol(node.val)

        if symbol_val is None:
            err_msg = 'Variable "{}" is not defined!.'.format(node.val)
            raise InternalRuntimeErr(err_msg)
        else:
            return symbol_val

    def visit_string(self, node):
        return String_Object(node.val)

    def visit_bracket(self, node):
        return node.expr.accept(self)

    def visit_bool(self, node):
        if node.val == Tok_Type.TRUE:
            return Bool_Object(True)
        else:
            return Bool_Object(False)

    def visit_arith(self, node):
        # Evaluate left and right sides
        left = node.left.accept(self)
        right = node.right.accept(self)

        # Perform actions depeding on type
        if node.op == Tok_Type.PLUS:
            return left + right
        elif node.op == Tok_Type.MINUS:
            return left - right
        elif node.op == Tok_Type.MUL:
            return left * right
        elif node.op == Tok_Type.DIV:
            return left / right

    def visit_logic(self, node):
        left = node.left.accept(self)
        right = node.right.accept(self)

        # >
        if node.op == Tok_Type.AND:
            return left & right
        # >=
        elif node.op == Tok_Type.OR:
            return left | right

    def visit_relop(self, node):
        left = node.left.accept(self)
        right = node.right.accept(self)

        # >
        if node.op == Tok_Type.GT:
            return left > right
        # >=
        elif node.op == Tok_Type.GTE:
            return left >= right
        # <
        elif node.op == Tok_Type.LT:
            return left < right
        # <=
        elif node.op == Tok_Type.LTE:
            return left <= right

    def visit_eq(self, node):
        left = node.left.accept(self)
        right = node.right.accept(self)

        if node.op == Tok_Type.EQ:
            return left == right
        elif node.op == Tok_Type.NEQ:
            return left != right

    def visit_unary(self, node):
        result = node.value.accept(self)
        if node.op == Tok_Type.MINUS:
            return -result
        elif node.op == Tok_Type.BANG:
            return ~result

    def visit_if(self, node):
        # Evaluate each condition and execute block if it is true
        for cond_block in node.cond_blocks:
            cond_value = cond_block["cond"].accept(self)
            if is_truthy(cond_value):
                return cond_block["block"].accept(self)

        if node.else_block is not None:
            return node.else_block.accept(self)

    def visit_while(self, node):
        cond_value = node.condition.accept(self)
        while is_truthy(cond_value):
            # Detect any usage of return
            ret_val = node.body.accept(self)
            if ret_val is not None:
                # Exit loop early using return
                if ret_val.obj_type == ObjectType.BREAK:
                    return
                # There is no need to do anything here. We need to reevaluate the test
                # condition for the loop.
                elif ret_val.obj_type == ObjectType.CONTINUE:
                    continue
                # "return" value is floated up
                else:
                    return ret_val
            cond_value = node.condition.accept(self)

    def visit_print(self, node):
        print_value = node.expr.accept(self)
        print(str(print_value))

    def visit_replace_assign(self, node):
        symbol_namespace = None
        symbol_name = None
        new_value = node.value.accept(self)
        if node.symbol_path is not None:
            symbol_namespace = node.symbol_path.accept(self)
        else:
            symbol_namespace = self.env.peek()

        if isinstance(node.symbol_name, Array_Access_Node):
            symbol_name = node.symbol_name.array_name.val
            array_index = node.symbol_name.array_pos.accept(self)
            if array_index.obj_type != ObjectType.NUM:
                err_msg = 'Array cannot be "{}" !'.format(array_index.obj_type)
                raise InternalRuntimeErr(err_msg)

            array_instance = symbol_namespace.lookup_symbol(symbol_name)
            if array_instance is None:
                err_msg = (
                    'The array "{}" does not exist within the current' "environment!"
                ).format(symbol_name)
                raise InternalRuntimeErr(err_msg)
            else:
                if array_index.value < array_instance.size:
                    array_instance.elements[array_index.value] = new_value
                else:
                    err_msg = '"{}" exceeds the length of the array "{}"!'.format(
                        array_index.value, symbol_name
                    )
                    raise InternalRuntimeErr(err_msg)
        else:
            symbol_name = node.symbol_name.val
            if isinstance(symbol_namespace, Scope):
                status = symbol_namespace.replace_variable(symbol_name, new_value)
                if status is False:
                    err_msg = (
                        'Variable "{}" cannot be reasigned because it has not'
                        "been initialized!"
                    ).format(symbol_name)
                    raise InternalRuntimeErr(err_msg)
            elif symbol_namespace.obj_type in [
                ObjectType.MODULE,
                ObjectType.CLASS_INSTANCE,
            ]:
                status = symbol_namespace.namespace.replace_variable(
                    symbol_name, new_value
                )
                if status is False:
                    err_msg = (
                        'Variable "{}" cannot be reasigned because it has not'
                        "been initialized!"
                    ).format(symbol_name)
                    raise InternalRuntimeErr(err_msg)

    def visit_new_assign(self, node):
        # Symbol name will always be an id node
        symbol_name = node.symbol_name.val
        # Evaluate the right hand side containing the value which will be assigned
        value = node.value.accept(self)

        assert symbol_name is not None, (
            "Attempting to assign a new variable without a symbol name in"
            "visit_new_assign."
        )
        assert (
            value is not None
        ), "Attempting to assign a non existant value in visit_new_assign."

        # Find if a path to the variable exists
        symbol_path = None
        if node.symbol_path is None:
            # We must be trying to access a variable within the current scope.
            scope = self.env.peek()
            scope.new_variable(symbol_name, value)
        else:
            symbol_path = node.symbol_path.accept(self)
            if isinstance(symbol_path, Scope):
                symbol_path.new_variable(symbol_name, value)
            elif symbol_path.obj_type in [ObjectType.MODULE, ObjectType.CLASS_INSTANCE]:
                symbol_path.namespace.new_variable(symbol_name, value)

    def visit_scope_block(self, node):
        # Create a new scope to evaluate the current block in
        parent_env = self.env.peek()
        self.env.enter_scope(parent_env)
        for stmnt in node.stmnts:
            ret_val = stmnt.accept(self)
            # Bubble up any flow statements
            if ret_val is not None and ret_val.obj_type in [
                ObjectType.RETURN,
                ObjectType.BREAK,
                ObjectType.CONTINUE,
            ]:
                self.env.exit_scope()
                return ret_val

        self.env.exit_scope()

    def visit_switch(self, node):
        # Evaluate the condition used for testing all cases
        test_cond = node.switch_cond.accept(self)

        # Find the index of the first switch case which is true.
        start_case_idx = None
        for idx, switch_case in enumerate(node.switch_cases):
            case_cond = switch_case[0].accept(self)
            if is_truthy(case_cond == test_cond):
                start_case_idx = idx
                break

        # Execute all switch cases after that until we encounter a "break" keyword
        # or a "return"/"continue" keywords.
        for _, case_body in node.switch_cases[start_case_idx:]:
            ret_val = case_body.accept(self)
            if ret_val is not None:
                if ret_val.obj_type == ObjectType.BREAK:
                    return
                else:
                    return ret_val

        # Execute the default case if it is provided.
        if node.default_case is not None:
            return node.default_case.accept(self)

    def visit_func_def(self, node):
        curr_scope = self.env.peek()
        # Register the function in the current frame
        curr_scope.new_variable(
            node.name, Func_Object(node.name, node.args, node.body, curr_scope)
        )

    def visit_class_def(self, node):
        """
        Keyword Arguments:
        node -- Class definition node
        """
        curr_scope = self.env.peek()
        # Register the class in the scope
        curr_scope.new_variable(
            node.class_name, Class_Def_Object(node.class_name, node.class_methods)
        )

    def __run_native_function(self, func_object, call_args):
        if func_object.arity != len(call_args):
            raise InternalRuntimeErr(
                "Function '{}' accepts only {} arguments but {} were given".format(
                    func_object.name, func_object.arity, len(call_args)
                )
            )

        evaluated_args = list()
        for arg in call_args:
            val = arg.accept(self)
            evaluated_args.append(val)

        # Using the "*" operator will destructure the list of evaluated internal object
        # arguments into the arguments which the function accepts.
        return func_object.body(*evaluated_args)

    def __run_internal_function(self, func_object, call_args):
        """
        Runs the function represented by func_object. The arguments passed are supplied
        by the call_args in the form of a list.
        """
        # Evaluate the arguments
        arg_values = list()
        for arg in call_args:
            val = arg.accept(self)
            arg_values.append(val)

        if len(arg_values) != func_object.arity:
            msg = 'function "{}" accepts only {} arguments but {} were given!'.format(
                func_object.name,
                func_object.arity,
                len(arg_values),
            )
            raise InternalRuntimeErr(msg)

        if func_object.obj_type == ObjectType.FUNC:
            # Create a new scope
            self.env.enter_scope(func_object.env)
        elif func_object.obj_type == ObjectType.CLASS_METHOD:
            self.env.enter_scope(func_object.class_env)
            self.env.peek().new_variable("this", func_object.class_env)

        # Add arguments to the current environment
        for arg_pair in zip(func_object.args, arg_values):
            # print("Adding pair to environment", arg_pair)
            self.env.peek().new_variable(arg_pair[0].lexeme, arg_pair[1])

        for stmnt in func_object.body:
            ret_val = stmnt.accept(self)
            if ret_val is not None and ret_val.obj_type in [
                ObjectType.RETURN,
                ObjectType.BREAK,
                ObjectType.CONTINUE,
            ]:
                if ret_val.obj_type == ObjectType.RETURN:
                    return ret_val
                elif ret_val.obj_type == ObjectType.BREAK:
                    msg = '"break" statement not used within a loop or a switch block!'
                    raise InternalRuntimeErr(msg)
                elif ret_val.obj_type == ObjectType.CONTINUE:
                    msg = '"continue" statement not used within a loop!'
                    raise InternalRuntimeErr(msg)
        return None

    def visit_call(self, node):
        call_object = node.object_name.accept(self)

        # Case of function object
        if call_object.obj_type in [ObjectType.FUNC, ObjectType.CLASS_METHOD]:
            ret_val = self.__run_internal_function(call_object, node.call_args)
            self.env.exit_scope()
            if ret_val is None or ret_val.value is None:
                return Nil_Object()
            else:
                return ret_val.value

        elif call_object.obj_type == ObjectType.NATIVE_FUNC:
            return self.__run_native_function(call_object, node.call_args)

        elif call_object.obj_type == ObjectType.CLASS_DEF:
            instance_ptr = Class_Instance_Object(
                call_object.class_name, call_object.class_methods
            )
            # Enter new scope to register "this" namespace
            self.env.enter_scope(instance_ptr.namespace)
            self.env.peek().new_variable("this", instance_ptr.namespace)

            class_constructor = instance_ptr.get_field("constructor")
            if class_constructor is None and len(node.call_args) != 0:
                raise InternalRuntimeErr(
                    (
                        "Class '{}' does not have a constructor method but "
                        "initialization detected {} arguments passed."
                    ).format(call_object.class_name, len(node.call_args))
                )
            elif class_constructor is not None:
                self.__run_internal_function(class_constructor, node.call_args)

            self.env.exit_scope()
            return instance_ptr
        else:
            raise InternalRuntimeErr("Object is not callable!")

    def visit_dot_node(self, node):
        left_side = node.left.accept(self)
        if isinstance(left_side, Scope):
            val = left_side.lookup_symbol(node.right.val)
            if val is not None:
                return val
            else:
                err_msg = "Current environment does not contain the variable {}".format(
                    node.right.val
                )
                raise InternalRuntimeErr(err_msg)
        elif left_side.obj_type == ObjectType.MODULE:
            val = left_side.namespace.lookup_symbol(node.right.val)
            if val is not None:
                return val
            else:
                err_msg = "Module environment does not contain the variable {}".format(
                    node.right.lexeme
                )
                raise InternalRuntimeErr(err_msg)
        elif left_side.obj_type == ObjectType.CLASS_INSTANCE:
            val = left_side.get_field(node.right.val)
            if val is not None:
                return val
            else:
                err_msg = (
                    'Class instance "{}" of class "{}" does not contain a field '
                    'with name "{}"'
                ).format(node.left.val, left_side.class_name, node.right.val)
                raise InternalRuntimeErr(err_msg)
        else:
            err_msg = "variable {} is not accessible!".format(node.left.val)
            raise InternalRuntimeErr(err_msg)

    def visit_this(self, node):
        curr_scope = self.env.peek()
        while curr_scope.parent is not None:
            curr_scope = curr_scope.parent
        return curr_scope

    def visit_return(self, node):
        if node.expr is None:
            return Return_Object(Nil_Object())

        return_val = node.expr.accept(self)
        return Return_Object(return_val)

    def visit_continue(self, node):
        return Continue_Object()

    def visit_break(self, node):
        return Break_Object()

    def visit_do_while(self, node):
        # First execution of the body which always happens
        ret_val = node.body.accept(self)
        if ret_val is not None:
            # Exit loop early using return
            if ret_val.obj_type == ObjectType.BREAK:
                return
            # There is no need to do anything here. We need to reevaluate the test
            # condition for the loop.
            elif ret_val.obj_type == ObjectType.CONTINUE:
                pass
            # "return" value is floated up
            else:
                return ret_val

        # Subsequent executions which depend on the ocndition
        cond_value = node.cond.accept(self)
        while is_truthy(cond_value):
            # Detect any usage of return
            ret_val = node.body.accept(self)
            if ret_val is not None:
                # Exit loop early using return
                if ret_val.obj_type == ObjectType.BREAK:
                    return
                # There is no need to do anything here. We need to reevaluate the test
                # condition for the loop.
                elif ret_val.obj_type == ObjectType.CONTINUE:
                    continue
                # "return" value is floated up
                else:
                    return ret_val
            cond_value = node.cond.accept(self)

    def visit_array(self, node):
        eval_elem = list()
        for elem in node.elements:
            elem_val = elem.accept(self)
            eval_elem.append(elem_val)

        return Array_Object(eval_elem)

    def visit_array_access(self, node):
        array_obj = node.array_name.accept(self)
        array_idx = node.array_pos.accept(self)
        if array_obj.obj_type != ObjectType.ARRAY:
            err_str = "Object is not an array and cannot be accessed using '[]'!"
            raise InternalRuntimeErr(err_str)
        if array_idx.obj_type != ObjectType.NUM:
            err_str = "Array index is not a number but a '{}'!".format(
                str(array_idx.obj_type)
            )
            raise InternalRuntimeErr(err_str)
        if array_idx.value < array_obj.size:
            return array_obj.elements[array_idx.value]
        else:
            msg = "Array has a size of {} but you want to access position {}".format(
                array_obj.size, array_idx.value
            )
            raise InternalRuntimeErr(msg)

    def visit_incr(self, node):
        node_val = node.value.accept(self)
        node_val.value += 1
        return node_val

    def visit_decr(self, node):
        node_val = node.value.accept(self)
        node_val.value -= 1
        return node_val

    def visit_nil(self, node):
        return Nil_Object()

    def visit_import(self, node):
        module_name = node.module_name + ".zai"
        module_path, module_text = read_module_contents(module_name)
        if module_text is None:
            err_msg = (
                "Module {} could not be found within the interpreter path.".format(
                    node.filename
                )
            )
            raise InternalRuntimeErr(err_msg)

        # Initialize lexer and environment used to execute module contents
        lexer = Lexer()
        import_env = Environment()

        # Lex and parse module contents
        tok_stream = lexer.tokenize_string(module_text)
        parser = Parser(tok_stream, module_text)
        root = parser.parse()

        import_visitor = Visitor(import_env)
        import_visitor.visit(root)
        # Take the module's global environment to be added to the namespace.
        import_scope = import_visitor.env.peek()

        # Determine the name with which the module will be accessed.
        module_env_name = node.module_name
        if node.import_name is not None:
            module_env_name = node.import_name

        self.env.peek().new_variable(
            module_env_name,
            Module_Object(node.module_name, module_path, import_scope, module_env_name),
        )

    def visit_add_assign(self, node):
        # Symbol name will always be an id node
        symbol_name = node.symbol_name.val
        # Evaluate the right hand side containing the value which will be assigned
        new_value = node.increment.accept(self)

        # Find if a path to the variable exists
        symbol_path = None
        if node.symbol_path is None:
            current_scope = self.env.peek()
            old_val = current_scope.lookup_symbol(symbol_name)
            status = current_scope.replace_variable(symbol_name, old_val + new_value)
            if status is False:
                err_msg = (
                    'Variable "{}" cannot be reasigned because it does not exist'
                    " within the environment."
                ).format(symbol_name)
                raise InternalRuntimeErr(err_msg)
        else:
            symbol_path = node.symbol_path.accept(self)
            if isinstance(symbol_path, Scope):
                old_val = symbol_path.lookup_symbol(symbol_name)
                status = symbol_path.replace_variable(symbol_name, old_val + new_value)
                if status is False:
                    err_msg = (
                        'Variable "{}" cannot be reasigned because it does not'
                        "exist within the current class environment."
                    ).format(symbol_name)
                    raise InternalRuntimeErr(err_msg)
            elif symbol_path.obj_type in [ObjectType.MODULE, ObjectType.CLASS_INSTANCE]:
                old_val = symbol_path.namespace.lookup_symbol(symbol_name)
                status = symbol_path.namespace.replace_variable(
                    symbol_name, old_val + new_value
                )
                if status is False:
                    err_msg = (
                        'Variable "{}" cannot be reasigned because it does not '
                        "exist."
                    ).format(symbol_name)
                    raise InternalRuntimeErr(err_msg)

    def visit_sub_assign(self, node):
        # Symbol name will always be an id node
        symbol_name = node.symbol_name.val
        # Evaluate the right hand side containing the value which will be assigned
        new_value = node.decrement.accept(self)

        # Find if a path to the variable exists
        symbol_path = None
        if node.symbol_path is None:
            current_scope = self.env.peek()
            old_val = current_scope.lookup_symbol(symbol_name)
            status = current_scope.replace_variable(symbol_name, old_val - new_value)
            if status is False:
                err_msg = (
                    'Variable "{}" cannot be reasigned because it does not exist'
                    " within the  environment."
                ).format(symbol_name)
                raise InternalRuntimeErr(err_msg)
        else:
            symbol_path = node.symbol_path.accept(self)
            if isinstance(symbol_path, Scope):
                old_val = symbol_path.lookup_symbol(symbol_name)
                status = symbol_path.replace_variable(symbol_name, old_val - new_value)
                if status is False:
                    err_msg = (
                        'Variable "{}" cannot be reasigned because it does not '
                        "exist within the current class environment."
                    ).format(symbol_name)
                    raise InternalRuntimeErr(err_msg)
            elif symbol_path.obj_type in [ObjectType.MODULE, ObjectType.CLASS_INSTANCE]:
                old_val = symbol_path.namespace.lookup_symbol(symbol_name)
                status = symbol_path.namespace.replace_variable(
                    symbol_name, old_val - new_value
                )
                if status is False:
                    err_msg = (
                        'Variable "{}" cannot be reasigned because it does not '
                        "exist."
                    ).format(symbol_name)
                    raise InternalRuntimeErr(err_msg)



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

"""Module containing utilities functions used by the interpreter."""
import os
from zai.objects import ObjectType


def get_module_path():
    """
    Retrieve the interpreter's module path from the environment.
    """
    # retrieve the current working directory and the interpreter path specified
    # in the environment
    curr_path = [os.getcwd()]
    environ_path = os.environ.get("ZAI_PATH").split(":")

    # Remove any strings containing only whitespace and combine with current working
    # directory path.
    environ_path = list(filter(None, environ_path))
    return curr_path + environ_path


def read_module_contents(module_name):
    """
    Find and return a string containing the contents of a module.
    """
    from os import listdir

    module_path = get_module_path()
    full_module_path = None
    for path in module_path:
        if module_name in listdir(path):
            full_module_path = os.path.join(path, module_name)
            return (full_module_path, open(full_module_path, "r").read())
    return (None, None)


def is_truthy(internal_object):
    """ Check if an internal object is truthy. Returns True or False."""
    # Truthiness will be the same as the one in python
    if internal_object.obj_type in [
        ObjectType.BOOL,
        ObjectType.STR,
        ObjectType.NUM,
        ObjectType.NIL,
        ObjectType.ARRAY,
    ]:
        return bool(internal_object)
    else:
        # Functions, class instances and class definitions are not truthy.
        return False


def is_atom(obj):
    """
    Check if an object is an atom(Boolean, Integer or string).
    """
    if obj is None:
        return False
    return obj.obj_type in [
        ObjectType.BOOL,
        ObjectType.NUM,
        ObjectType.STR,
        ObjectType.NIL,
    ]



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

""" Module defining nodes used in the abstract syntax tree created by the parser. """


class AST_Node:
    """
    Base class from which all AST nodes are derived.
    """

    def __init__(self):
        raise NotImplementedError()


class Program_Node(AST_Node):
    def __init__(self):
        self.stmnts = list()

    def add_stmnt(self, stmnt):
        self.stmnts.append(stmnt)

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_program(self)

    def __str__(self):
        # print(self.stmnts)
        for stmn in self.stmnts:
            print(stmn)
        return "Program_node"


class Print_Node(AST_Node):
    def __init__(self, expr):
        self.expr = expr

    def accept(self, visitor):
        visitor.visit_print(self)

    def __str__(self):
        return "{}".format(self.expr)


class Bin_Node(AST_Node):
    def __init__(self, left, op, right):
        raise NotImplementedError()


class Dot_Bin_Node(AST_Node):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __str__(self):
        return "DOT_NODE {} {}".format(self.left, self.right)

    def __repr__(self):
        return "DOT_NODE {} {}".format(self.left, self.right)

    def accept(self, visitor):
        return visitor.visit_dot_node(self)


class Replace_Assign_Bin_Node(AST_Node):
    def __init__(self, symbol_path, symbol_name, value):
        self.symbol_path = symbol_path  # Path leading to the symbol
        self.symbol_name = symbol_name  # The actualy symbol name within the environment
        self.value = value

    def __str__(self):
        return "REPLACE_ASSIGN_NODE {} {} {}".format(
            self.symbol_name, self.symbol_path, self.value
        )

    def accept(self, visitor):
        return visitor.visit_replace_assign(self)


class New_Assign_Bin_Node(AST_Node):
    def __init__(self, symbol_path, symbol_name, value):
        self.symbol_path = symbol_path
        self.symbol_name = symbol_name
        self.value = value

    def __str__(self):
        return "NEW_ASSIGN_NODE {} {} {}".format(
            self.symbol_name, self.symbol_path, self.value
        )

    def accept(self, visitor):
        return visitor.visit_new_assign(self)


class Eq_Bin_Node(Bin_Node):
    def __init__(self, left, op, right):
        self.left = left
        self.right = right
        self.op = op

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_eq(self)

    def __str__(self):
        return "EQ_NODE: {} {} {}".format(self.left, self.op, self.right)


class Arith_Bin_Node(Bin_Node):
    def __init__(self, left, op, right):
        self.left = left
        self.right = right
        self.op = op

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_arith(self)

    def __str__(self):
        return "ARITH_NODE: {} {} {}".format(self.left, self.op, self.right)


class Logic_Bin_Node(Bin_Node):
    def __init__(self, left, op, right):
        self.left = left
        self.right = right
        self.op = op

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_logic(self)

    def __str__(self):
        return "LOGIC_NODE: {} {} {}".format(self.left, self.op, self.right)


class Relop_Bin_Node(Bin_Node):
    def __init__(self, left, op, right):
        self.left = left
        self.right = right
        self.op = op

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_relop(self)

    def __str__(self):
        return "RELOP_NODE: {} {} {}".format(self.left, self.op, self.right)


class Unary_Node(AST_Node):
    def __init__(self, op, right):
        self.value = right
        self.op = op

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_unary(self)

    def __str__(self):
        return "UNARY_NODE: {} {}".format(self.op, self.right)


class Bracket_Node(AST_Node):
    def __init__(self, expr):
        self.expr = expr

    def __str__(self):
        return "BRACKET_NODE: {}".format(self.expr)

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_bracket(self)


class If_Node(AST_Node):
    def __init__(self, conditions, else_block):
        self.cond_blocks = conditions
        self.else_block = else_block

    def __str__(self):
        return "IF_NODE: conditions: {} else:{}".format(
            self.conditions, self.else_block
        )

    def accept(self, visitor):
        return visitor.visit_if(self)


class While_Node(AST_Node):
    def __init__(self, condition, body):
        self.condition = condition
        self.body = body

    def __str__(self):
        return "WHILE_NODE: condition: {} body:{}".format(self.condition, self.body)

    def accept(self, visitor):
        return visitor.visit_while(self)


class Switch_Node(AST_Node):
    def __init__(self, switch_cond, switch_cases, default_case):
        self.switch_cond = switch_cond
        self.switch_cases = switch_cases
        self.default_case = default_case

    def __str__(self):
        return "SWITCH_NODE: condition: {} cases:{} default:{}".format(
            self.switch_cond, self.switch_cases, self.default_case
        )

    def accept(self, visitor):
        return visitor.visit_switch(self)


class ID_Node(AST_Node):
    def __init__(self, node_val):
        self.val = node_val

    def __str__(self):
        return "ID_NODE: {}".format(self.val)

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_symbol(self)


class Bool_Node(AST_Node):
    def __init__(self, node_val):
        self.val = node_val

    def __str__(self):
        return "BOOL_NODE: {}".format(self.val)

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_bool(self)


class String_Node(AST_Node):
    def __init__(self, node_val):
        self.val = node_val

    def __str__(self):
        return "STR_NODE: {}".format(self.val)

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_string(self)


class Num_Node(AST_Node):
    def __init__(self, node_val):
        self.val = node_val

    def __str__(self):
        return "NUM_NODE: {}".format(self.val)

    def accept(self, visitor):
        """
        Keyword Arguments:
        visitor -- The visitor instance used to evaluate the node.
        """
        return visitor.visit_num(self)


class Func_Node(AST_Node):
    def __init__(self, name, args, body):
        self.name = name
        self.args = args
        self.body = body

    def __str__(self):
        return "FUNC_NODE {}".format(self.name)

    def accept(self, visitor):
        return visitor.visit_func_def(self)


class Scope_Block_Node(AST_Node):
    def __init__(self, block_stmnts):
        self.stmnts = block_stmnts

    def __str__(self):
        output = str()
        for stmtn in self.stmnts:
            output += "\n" + stmtn.__str__()
        return "BLOCK_NODE \n{}".format(output)

    def accept(self, visitor):
        return visitor.visit_scope_block(self)


class Call_Node(AST_Node):
    def __init__(self, object_name, call_args):
        self.object_name = object_name
        self.call_args = call_args

    def __str__(self):
        return "CALL_NODE: {} {}".format(self.object_name, self.call_args)

    def accept(self, visitor):
        return visitor.visit_call(self)


class Class_Method_Node(AST_Node):
    def __init__(self, name, args, body):
        self.name = name
        self.args = args
        self.body = body

    def __str__(self):
        return "CLASS_METHOD_NODE {}".format(self.name)

    def accept(self, visitor):
        return visitor.visit_class_method(self)


class Class_Def_Node(AST_Node):
    def __init__(self, class_name, class_methods):
        self.class_name = class_name
        self.class_methods = class_methods

    def __str__(self):
        return "CLASS_NODE {} {}".format(self.class_name, self.class_methods)

    def accept(self, visitor):
        return visitor.visit_class_def(self)


class This_Node(AST_Node):
    def __init__(self):
        pass

    def __str__(self):
        return "THIS_NODE"

    def accept(self, visitor):
        return visitor.visit_this(self)


class Break_Node(AST_Node):
    def __init__(self):
        pass

    def __str__(self):
        return "BREAK_NODE"

    def accept(self, visitor):
        return visitor.visit_break(self)


class Continue_Node(AST_Node):
    def __init__(self):
        pass

    def __str__(self):
        return "CONTINUE_NODE"

    def accept(self, visitor):
        return visitor.visit_continue(self)


class Return_Node(AST_Node):
    def __init__(self, return_expr):
        self.expr = return_expr

    def __str__(self):
        return "RETURN_NODE {}".format(self.expr)

    def accept(self, visitor):
        return visitor.visit_return(self)


class Do_While_Node(AST_Node):
    def __init__(self, cond, body):
        self.cond = cond
        self.body = body

    def __str__(self):
        return "DO_WHILE_NODE {}, {}".format(self.cond, self.body)

    def accept(self, visitor):
        return visitor.visit_do_while(self)


class Array_Node(AST_Node):
    def __init__(self, elements):
        # elements is a list of "or_expr"
        self.elements = elements

    def __str__(self):
        return "ARRAY_NODE elements: {}".format(self.elements)

    def accept(self, visitor):
        return visitor.visit_array(self)


class Array_Access_Node(AST_Node):
    def __init__(self, array_id, array_position):
        self.array_name = array_id
        self.array_pos = array_position

    def __str__(self):
        return "ARRAY_ACCESS_NODE name: {}, position: {}".format(
            self.array_name, self.array_pos
        )

    def accept(self, visitor):
        return visitor.visit_array_access(self)


class Incr_Node(AST_Node):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "INCR_NODE {}".format(self.value)

    def accept(self, visitor):
        return visitor.visit_incr(self)


class Decr_Node(AST_Node):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "DECR_NODE {}".format(self.value)

    def accept(self, visitor):
        return visitor.visit_decr(self)


class Nil_Node(AST_Node):
    def __init__(self):
        pass

    def __str__(self):
        return "NIL_NODE"

    def accept(self, visitor):
        return visitor.visit_nil(self)


class Import_Node(AST_Node):
    def __init__(self, module_name, import_name=None):
        self.module_name = module_name
        self.import_name = import_name

    def __str__(self):
        return "IMPORT_NODE {} with import name: {}".format(
            self.module_name, self.import_name
        )

    def accept(self, visitor):
        return visitor.visit_import(self)


class AddAssign_Node(AST_Node):
    def __init__(self, symbol_path, symbol_name, increment):
        self.symbol_path = symbol_path
        self.symbol_name = symbol_name
        self.increment = increment

    def __str__(self):
        return "ADD_ASSIGN_NODE {} {} {}".format(
            self.symbol_name, self.symbol_path, self.increment
        )

    def accept(self, visitor):
        return visitor.visit_add_assign(self)


class SubAssign_Node(AST_Node):
    def __init__(self, symbol_path, symbol_name, decrement):
        self.symbol_path = symbol_path
        self.symbol_name = symbol_name
        self.decrement = decrement

    def __str__(self):
        return "SUB_ASSIGN_NODE {} {} {}".format(
            self.symbol_name, self.symbol_path, self.decrement
        )

    def accept(self, visitor):
        return visitor.visit_sub_assign(self)



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

"""
Module containing lexer class used to convert an input string into a sequence
of language tokens.
"""
from zai.tokens import Tok_Type, Token, keywords
from zai.internal_error import InternalTokenErr


class Lexer:
    """Class used to convert a string of characters into language tokens."""

    def __init__(self):
        self.text = None
        self.token_stream = list()
        # Current char observed
        self.curr_char = None
        # Current index within the input stream
        self.curr_idx = -1
        # Total length of input stream
        self.input_len = 0

        self.curr_lin_num = 0
        self.curr_col_num = 0

        # Special characters which are restricted/permitted in ID tokens.
        self.restricted_ident_chars = ".,:;()|&[]*/+-<=>!{}#\"'\n\t "
        self.permitted_ident_chars = "?@$"

    def _advance(self):
        """Advance the current character by one and return it. If there is no next character,
        return None."""
        if self.curr_idx + 1 < self.input_len:
            self.curr_idx += 1
            self.curr_char = self.text[self.curr_idx]
            self.curr_col_num += 1
            return self.curr_char
        else:
            self.curr_char = None
            return None

    def _peek(self):
        """Return the next character in the input text sequence. If there is no next
        character, return None."""
        if self.curr_idx + 1 < self.input_len:
            return self.text[self.curr_idx + 1]
        else:
            return None

    def _tokenize_ident(self):
        """ Tokenize an identifier and return its lexeme."""
        ident_str = str(self.curr_char)
        # ident_str += self.curr_char
        while (
            self._peek() is not None and self._peek() not in self.restricted_ident_chars
        ):
            self._advance()
            ident_str += self.curr_char
        return ident_str

    def _tokenize_num(self):
        """ Tokenize a single integer number and return it in integer form."""
        num_str = str()
        num_str += self.curr_char
        while self._peek() is not None and self._peek() in "1234567890":
            self._advance()
            num_str += self.curr_char
        return num_str

    def _tokenize_str(self):
        str_content = str()
        while True:
            if self.curr_char is None:
                self.token_stream.append(
                    Token(
                        Tok_Type.STRING,
                        str_content,
                        self.curr_lin_num,
                        self.curr_col_num,
                    )
                )
                return
            elif self.curr_char != "\\" and self._peek() == '"':
                str_content += self.curr_char
                self.token_stream.append(
                    Token(
                        Tok_Type.STRING,
                        str_content,
                        self.curr_lin_num,
                        self.curr_col_num,
                    )
                )
                self._advance()
                self.token_stream.append(
                    Token(Tok_Type.DQUOTE, None, self.curr_lin_num, self.curr_col_num)
                )
                return
            else:
                str_content += self.curr_char
                self._advance()

    def _tokenize(self):
        """ Tokenize the current text sequence and return the tokens generated. """
        while self._advance() is not None:
            if self.curr_char == "\n":
                # Increment line numbers
                self.curr_lin_num += 1
                # Reset current column number
                self.curr_col_num = 0
            elif self.curr_char == "(":
                self.token_stream.append(
                    Token(Tok_Type.LROUND, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == ")":
                self.token_stream.append(
                    Token(Tok_Type.RROUND, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == ".":
                self.token_stream.append(
                    Token(Tok_Type.DOT, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == ",":
                self.token_stream.append(
                    Token(Tok_Type.COMMA, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == ":":
                self.token_stream.append(
                    Token(Tok_Type.COLON, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == ";":
                self.token_stream.append(
                    Token(Tok_Type.SEMIC, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == "{":
                self.token_stream.append(
                    Token(Tok_Type.LCURLY, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == "}":
                self.token_stream.append(
                    Token(Tok_Type.RCURLY, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == "[":
                self.token_stream.append(
                    Token(Tok_Type.LSQUARE, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == "]":
                self.token_stream.append(
                    Token(Tok_Type.RSQUARE, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == "-":
                if self._peek() == "-":
                    self.token_stream.append(
                        Token(Tok_Type.DECR, None, self.curr_lin_num, self.curr_col_num)
                    )
                    self._advance()
                elif self._peek() == "=":
                    self.token_stream.append(
                        Token(
                            Tok_Type.SUBASSIGN,
                            None,
                            self.curr_lin_num,
                            self.curr_col_num,
                        )
                    )
                    self._advance()
                else:
                    self.token_stream.append(
                        Token(
                            Tok_Type.MINUS, None, self.curr_lin_num, self.curr_col_num
                        )
                    )
            elif self.curr_char == "&":
                if self._peek() == "&":
                    self.token_stream.append(
                        Token(Tok_Type.AND, None, self.curr_lin_num, self.curr_col_num)
                    )
                    self._advance()
                else:
                    raise InternalTokenErr(
                        self.curr_lin_num,
                        self.curr_col_num,
                        self.text,
                        (
                            "A single '&' is not a valid symbol or operator. Did you"
                            "mean to use '&&'(AND) operator?"
                        ),
                    )
            elif self.curr_char == "|":
                if self._peek() == "|":
                    self.token_stream.append(
                        Token(Tok_Type.OR, None, self.curr_lin_num, self.curr_col_num)
                    )
                    self._advance()
                else:
                    raise InternalTokenErr(
                        self.curr_lin_num,
                        self.curr_col_num,
                        self.text,
                        (
                            "A single '|' is not a valid symbol or operator. Did you "
                            "mean to use '||'(OR) operator?"
                        ),
                    )
            elif self.curr_char == "+":
                if self._peek() == "+":
                    self.token_stream.append(
                        Token(Tok_Type.INCR, None, self.curr_lin_num, self.curr_col_num)
                    )
                    self._advance()
                elif self._peek() == "=":
                    self.token_stream.append(
                        Token(
                            Tok_Type.ADDASSIGN,
                            None,
                            self.curr_lin_num,
                            self.curr_col_num,
                        )
                    )
                    self._advance()
                else:
                    self.token_stream.append(
                        Token(Tok_Type.PLUS, None, self.curr_lin_num, self.curr_col_num)
                    )
            elif self.curr_char == "/":
                self.token_stream.append(
                    Token(Tok_Type.DIV, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == "*":
                self.token_stream.append(
                    Token(Tok_Type.MUL, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == '"':
                self.token_stream.append(
                    Token(Tok_Type.DQUOTE, None, self.curr_lin_num, self.curr_col_num)
                )
                # Skip over the current character which is a double quote
                self._advance()
                self._tokenize_str()
            elif self.curr_char == "'":
                self.token_stream.append(
                    Token(Tok_Type.QUOTE, None, self.curr_lin_num, self.curr_col_num)
                )
            elif self.curr_char == "!":
                if self._peek() == "=":
                    self.token_stream.append(
                        Token(Tok_Type.NEQ, None, self.curr_lin_num, self.curr_col_num)
                    )
                    self._advance()
                else:
                    self.token_stream.append(
                        Token(Tok_Type.BANG, None, self.curr_lin_num, self.curr_col_num)
                    )
            elif self.curr_char == "=":
                if self._peek() == "=":
                    self.token_stream.append(
                        Token(Tok_Type.EQ, None, self.curr_lin_num, self.curr_col_num)
                    )
                    self._advance()
                else:
                    self.token_stream.append(
                        Token(
                            Tok_Type.ASSIGN, None, self.curr_lin_num, self.curr_col_num
                        )
                    )
            elif self.curr_char == "<":
                if self._peek() == "=":
                    self.token_stream.append(
                        Token(Tok_Type.LTE, None, self.curr_lin_num, self.curr_col_num)
                    )
                    self._advance()
                else:
                    self.token_stream.append(
                        Token(Tok_Type.LT, None, self.curr_lin_num, self.curr_col_num)
                    )
            elif self.curr_char == ">":
                if self._peek() == "=":
                    self.token_stream.append(
                        Token(Tok_Type.GTE, None, self.curr_lin_num, self.curr_col_num)
                    )
                    self._advance()
                else:
                    self.token_stream.append(
                        Token(Tok_Type.GT, None, self.curr_lin_num, self.curr_col_num)
                    )
            elif (
                self.curr_char in self.permitted_ident_chars or self.curr_char.isalpha()
            ):
                # Store the column at the start of the identifier
                ident_start_col = self.curr_col_num
                ident = self._tokenize_ident()

                # Check if the current identifier is a keyword.
                token = keywords.get(ident, None)
                if token is None:
                    token = Token(
                        Tok_Type.ID, ident, self.curr_lin_num, ident_start_col
                    )
                else:
                    token = Token(token, None, self.curr_lin_num, ident_start_col)

                self.token_stream.append(token)
            elif self.curr_char.isdigit():
                # Store the column at the start of the number
                num_start_col = self.curr_col_num
                num_str = self._tokenize_num()

                # Check if the number being tokenized might actually be a bad identifier
                # which starts with a number.
                # Example: 13abc or 1Alph@
                if self._peek() is not None and (
                    self._peek() == self.permitted_ident_chars or self._peek().isalpha()
                ):
                    raise InternalTokenErr(
                        self.curr_lin_num,
                        num_start_col,
                        self.text,
                        "Identifiers cannot start with integers!",
                    )
                self.token_stream.append(
                    Token(Tok_Type.NUM, int(num_str), self.curr_lin_num, num_start_col)
                )

        # Add final EOF token to indicate end of token stream
        self.token_stream.append(
            Token(Tok_Type.EOF, None, self.curr_lin_num, self.curr_col_num)
        )

    def _reset_internal_state(self):
        """Reset the internal state of the lexer for a new input string."""
        self.input_len = len(self.text)

        # Reset all other variables
        self.token_stream = list()
        self.curr_char = None
        self.curr_idx = -1
        self.curr_lin_num = 0
        self.curr_col_num = 0

    def tokenize_string(self, input_str):
        """ Tokenize a single input string. Returns the token stream produced."""
        if self.text is not None:
            self.text = input_str
            self._reset_internal_state()
        else:
            self.text = input_str
            self.input_len = len(input_str)

        if self.input_len > 0:
            self._tokenize()
            return self.token_stream
        else:
            return [Token(Tok_Type.EOF)]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zai-pl",
    version="0.8.1",
    author="Yavor Konstantinov",
    author_email="ykonstantinov1@gmail.com",
    description="A small programming language written for learning purposes.",
    license="GPLv3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        'console_scripts': ['zai-pl=zai.__main__:main'],
        },
    url="https://github.com/sehnsucht13/zai-pl",
    keywords=['zai', 'programming-language', 'zai-pl'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Programming Language :: Other",
        "Topic :: Software Development :: Interpreters",
    ],
    packages=['zai', 'zai/stdlib'],
    python_requires='>=3.6',
)



# Copyright 2021 by Yavor Konstantinov <ykonstantinov1@gmail.com>

# This file is part of zai-pl.

# zai-pl is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# zai-pl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with zai-pl. If not, see <https://www.gnu.org/licenses/>.

"""Module contains a class used to manage the entire virtual machine."""
from zai.lexer import Lexer
from zai.env import Environment
from zai.parse import Parser
from zai.visitor import Visitor
from zai.internal_error import (
    InternalRuntimeErr,
    InternalTypeError,
    InternalTokenErr,
    InternalParseErr,
)

import atexit
import os
import readline


class YAPL_VM:
    """
    Class representing a single instance of the YAPL virtual machine. Each command
    is evaluate within the same context.
    """

    def __init__(self):
        self.env = Environment()
        self.repl_mode_flag = False
        self.visitor = Visitor(self.env)
        self.__load_stdlib()
        self.current_completions = None

    def __load_stdlib(self):
        """
        Load both the standard library in the environment of the current VM instance.
        """
        from zai.stdlib.native_func import register_functions

        native_functions = register_functions()
        curr_scope = self.env.peek()
        for func in native_functions:
            curr_scope.new_variable(func.name, func)

    # def __complete(self, text, state):
    #     if state == 0:
    #         print("Text from complete", text)
    #         new_completions = list()
    #         curr_scope = self.env.peek()
    #         print(curr_scope.keys())
    #     elif state < len(a):
    #         return a[state]
    #     else:
    #         return None

    def __setup_readline(self):
        histfile = os.path.join(os.path.expanduser("~"), ".zai_history")
        try:
            readline.read_history_file(histfile)
            readline.set_history_length(2000)
        except FileNotFoundError:
            pass

        atexit.register(readline.write_history_file, histfile)

        # # Set up autocompletion with "TAB" key
        # readline.parse_and_bind("tab: complete")
        # readline.set_completer(self.__complete)

    def run_repl(self):
        """
        Start a REPL which evaluates every command provided within one VM context.
        """
        self.repl_mode_flag = True
        self.__load_stdlib()
        self.__setup_readline()
        while True:
            lexer = Lexer()
            try:
                str_input = input(">> ")
                tok_stream = lexer.tokenize_string(str_input)
                parser = Parser(tok_stream, str_input)
                root = parser.parse()
                val = self.visitor.visit(root)
                if val is not None:
                    print(str(val))
            except InternalRuntimeErr as e:
                print(e)
            except InternalTypeError as e:
                print(e)
            except InternalTokenErr as e:
                print(e)
            except InternalParseErr as e:
                print(e)

    def run_string(self, input_str):
        """
        Run a single string within the current VM context.
        """
        self.__load_stdlib()
        lexer = Lexer()
        try:
            tok_stream = lexer.tokenize_string(input_str)
            parser = Parser(tok_stream, input_str)
            root = parser.parse()
            self.visitor.visit(root)
        except InternalRuntimeErr as e:
            print(e)
        except InternalTypeError as e:
            print(e)
        except InternalTokenErr as e:
            print(e)
        except InternalParseErr as e:
            print(e)
